(* ::Package:: *)

newEntry="gameDef";
gameDef =
  "Game" ->
   <|
    "Name" -> "Game",
    "SuperClass"-> Null,
    
    "runTypeDefinition" :>Class[game,
		"$Properties"->{

			"currentState",
			"setPosition",
					
			"$Copy",
			
			"endStateQWinner",
			"endStateQ",
			
			"possibleMoves",
			
			"nextStateMoves",
			"nextPossibleStates",
			
			"makeMoves",

			"drawState",
			"gameGraphNodeShape",
			"gameGraph"

		},
		
		"$Init"[self_]:>(
			self["currentState"]=self["initialState"]),
		
		"setPosition"[self_]:>(Function[state,self["currentState"]=state]),
		
		"$Copy"[self_]:>Module[{originalState=self["currentState"],new},
			new=self["$Class"]["$New"];
			new["setPosition"][originalState];
			new
		],

		(** determine whether (current) state is an end-state, and if it is, specify winner **)
		"endStateQWinner"[self_]:>self["endStateQWinner"[self["currentState"]]],		
		
		(** determine whether (current) state is an end-state **)
		"endStateQ"[self_,state_]:>self["endStateQWinner"[state]]["endState"],
		"endStateQ"[self_]:>self["endStateQWinner"]["endState"],
										
		(** determine the list of all possible moves in (current) state **)
		"possibleMoves"[self_]:>self["possibleMoves"[self["currentState"]]],
		
		(** determine the next state from (current) state after making a list of moves **)
		"nextStateMoves"[self_,moves_,state_]:>Module[{},
			Fold[self["nextState"[#2,#1]]&,state, moves]
		],
		"nextStateMoves"[self_,moves_]:>self["nextStateMoves"[moves,self["currentState"]]],		

		(** return the list of all possible next states from (current) state  **)
		"nextPossibleStates"[self_,state_]:>Map[Function[self["nextStateMoves"[#,state]]],List/@self["possibleMoves"[state]]],
		"nextPossibleStates"[self_]:>self["nextPossibleStates"[self["currentState"]]],
		
		(** Making moves by updating current state from (current) state following a list of moves**)
		"makeMoves"[self_,moves_,state_]:>self["setPosition"][self["nextStateMoves"[moves,state]]],		
		"makeMoves"[self_,moves_]:>self["setPosition"][self["nextStateMoves"[moves,self["currentState"]]]],		
	
		(** Drawing **)
		
		(** draw a single position **)
		"drawState"[self_]:>self["drawState"[self["currentState"]]],
		
		(** draw a single position as part of a game graph **)
		"gameGraphNodeShape"[self_]:>Function[{coords,vertex,size},Inset[self["drawState"[vertex]],coords,Center,0.50]],	
		
		(** Generating game graphs **)
		"gameGraph"[self_,depth_,state_]:>ResourceFunction["FixedPointGraph"][
			self["nextPossibleStates"[#]]&,
			state,
			depth,
			VertexSize->0.7,
			VertexShapeFunction->self["gameGraphNodeShape"],
			GraphLayout->"LayeredDigraphEmbedding",
(*			PerformanceGoal->"Quality",*)
			ImageSize->900			
		],
		"gameGraph"[self_,depth_]:>self["gameGraph"[depth,self["currentState"]]],
		"gameGraph"[self_]:>self["gameGraph"[1]]
		
		],
     
    "createInstance" :> game@"$New",
    
    "Class Members" -> EntityList[{GameEntity["Combinatorial game"]}]
    |>;
