(* ::Package:: *)

newEntry="tictactoeDef";
tictactoeDef=
	"Tic-tac-toe"->
	<|
		"Name"->{"Tic-tac-toe"},
		"SuperClass"->GameEntity["m,n,k-game"],

		"runTypeDefinition":>
			Class[tictactoeGame ->mnkGame,
				"$Properties"->{
				
				"legalStateQ",
				"endStateQWinner",
				"possibleMoves",
				"nextState",
				"drawState"
			},
				
		"initialState"->
		<|
			"board"->ConstantArray[0,{3,3}],
			"toMove"->1
		|>,
		
		"legalStateQ"[self_]:>Function[state,Module[{},
			If[!AssociationQ[state],
				Print["Game state is not an association"];
				EntityUnregister[gameStore];
				Abort[];
			];
			
			If[Keys[state]!={"board","toMove"},
				Print["Game state has incorrect keys"];
				EntityUnregister[gameStore];
				Abort[];
			];
			
			If[!MemberQ[{1,2},state["toMove"]],
				Print["Player ",state["toMove"], " does not exist"];
				EntityUnregister[gameStore];
				Abort[];
			];
			
			If[!ListQ[state["board"]]||Dimensions[state["board"]]!={3,3}||!And@@(MemberQ[{0,1,2},#]&/@Flatten[state["board"]]),
				Print["Board state ",state["board"], " cannot be parsed."];
				EntityUnregister[gameStore];
				Abort[];
			];
		]],

		"endStateQWinner"[self_,state_]:>Module[
			{
				b=state["board"],
				rows,
				cols,
				diagNW,
				diagSW,
				endstate,
				wins1,
				wins2,
				winner,
				ans
			},
			rows=b;
			cols=Transpose@b;
			diagNW={Diagonal@b};
			diagSW={Diagonal@Reverse@b};
			wins1=ContainsAny[Join[rows,cols,diagNW,diagSW],{{1,1,1}}];
			wins2=ContainsAny[Join[rows,cols,diagNW,diagSW],{{2,2,2}}];
			If[wins1&&wins2,
				Print["In board state ",b, "both players are winners. This position could not arise through legal gameplay."];
				EntityUnregister[gameStore];
				Abort[];				
			];
			If[wins1&&!wins2,ans=<|"endState"->True,"winner"->1|>];
			If[!wins1&&wins2,ans=<|"endState"->True,"winner"->2|>];
			If[!wins1&&!wins2,ans=<|"endState"->False,"winner"->Null|>];
			ans
		],
		
		"possibleMoves"[self_,state_]:>Module[
		{g},
		g=self["endStateQ"[state]];
		If[
			g,
			{},
			Position[state["board"],_?((#==0)&)]]
		],

		"nextState"[self_,move_,state_]:>Module[{
			nextBoard=state["board"],
			player=state["toMove"]
		},
		
			If[!MemberQ[self["possibleMoves"[state],move]],
				Print["move not possible"];(*Message[nextState::impossibleMove]*)
				EntityUnregister[gameStore];
				Abort[];
			];

			nextBoard[[Sequence@@move]]=player;

			<|"board"->nextBoard,"toMove"->3-player|>

		],

		"drawState"[self_,state_]:>GraphicsGrid[state["board"]/.{0 -> Graphics[],
  			 1 -> Graphics[{Thickness -> 0.10, Style[Line[{{-1, -1}, {1, 1}}], Blue],Style[Line[{{-1, 1}, {1, -1}}], Blue]}, ImageSize -> {50,50},PlotRange -> {{-1.6, 1.6}, {-1.6, 1.6}}],
  			 2 -> Graphics[{Thickness -> 0.10, Style[Circle[{0,0},0.5],Red]}, ImageSize -> {50,50},PlotRange -> {{-0.75, 0.75}, {-0.75, 0.75}}]
  			 },
			 Frame -> All,
			 Spacings ->Scaled[0.01],
			 AspectRatio->1
			 ]
		],

		"createInstance" :> tictactoeGame@"$New",

		"Alternate names"->{""},
		"Description"->"Tic-tac-toe description",
		"Link"->Hyperlink["https://en.wikipedia.org/wiki/Tic-tac-toe"]
	|>;
